@@ -2,6 +2,7 @@
 The maui.utils model contains utility functions for multi-omics analysis
 using maui.
 """
+import keras
 import numpy as np
 import pandas as pd
 from scipy import stats
@@ -10,6 +11,7 @@ from scipy import spatial
 from scipy import cluster
 from collections import Counter
 from sklearn.svm import LinearSVC
+from sklearn.model_selection import KFold
 from sklearn.metrics import roc_curve, auc
 from sklearn.preprocessing import label_binarize
 from sklearn.preprocessing import StandardScaler
@@ -114,6 +116,28 @@ def filter_factors_by_r2(z, x, threshold=0.02):
     return z.iloc[:, pd.Series(scores)[pd.Series(scores) > threshold].index]
 
 
+def neural_path_weight_product(model):
+    """The product of weights along the path from from input
+    features to latent factors.
+
+    Parameters
+    ----------
+    model: a maui model
+
+    Returns
+    -------
+    nwp: (n_features, n_latent_factors) DataFrame of neural weight products
+    """
+    path_ws = list()
+    for layer in model.encoder.layers:
+        if isinstance(layer, keras.layers.core.Dense):
+            path_ws.append(layer.get_weights()[0])
+
+    return pd.DataFrame(
+        np.linalg.multi_dot(path_ws), index=model.x_.columns, columns=model.z_.columns
+    )
+
+
 def map_factors_to_feaures_using_linear_models(z, x):
     """Get feature <-> latent factors mapping from linear models.
     Runs one univariate (multi-output) linear model per latent factor in `z`,
@@ -183,7 +207,7 @@ def compute_roc(z, y, classifier=LinearSVC(C=0.001), cv_folds=10):
     Returns
     -------
     roc_curves: dict, one key per class as well as "mean", each value is a dataframe
-                containing the tpr (true positive rate) and fpr (falce positive rate)
+                containing the tpr (true positive rate) and fpr (false positive rate)
                 defining that class (or the mean) ROC.
     """
     class_names = sorted(y.unique())
@@ -427,9 +451,20 @@ def _cv_coxph_c(
     cph = lifelines.CoxPHFitter(penalizer=penalty)
     survdf = pd.concat([survival, z], axis=1, sort=False).dropna()
 
-    scores = lifelines.utils.k_fold_cross_validation(
-        cph, survdf, duration_column, event_col=observed_column, k=cv_folds
-    )
+    kfold = KFold(cv_folds)
+    scores = list()
+
+    for train_index, test_index in kfold.split(survdf):
+        x_train, x_test = survdf.iloc[train_index], survdf.iloc[test_index]
+
+        cph.fit(x_train, duration_column, observed_column)
+        cindex = lifelines.utils.concordance_index(
+            x_test[duration_column],
+            -cph.predict_partial_hazard(x_test),
+            x_test[observed_column],
+        )
+        scores.append(cindex)
+
     return scores
 
 
