@@ -7,6 +7,7 @@ import numpy as np
 import pandas as pd
 
 from maui import Maui
+from maui.maui_warnings import MauiWarning
 
 
 samples = [f"Sample_{i}" for i in range(10)]
@@ -67,7 +68,9 @@ def test_dict2array():
 def test_maui_saves_feature_correlations():
     maui_model = Maui(n_hidden=[10], n_latent=2, epochs=1)
     z = maui_model.fit_transform({"d1": df1, "d2": df2})
-    assert hasattr(maui_model, "feature_correlations")
+    r = maui_model.get_feature_correlations()
+    assert r is not None
+    assert hasattr(maui_model, "feature_correlations_")
 
 
 def test_maui_saves_w():
@@ -78,6 +81,36 @@ def test_maui_saves_w():
     assert hasattr(maui_model, "w_")
 
 
+def test_maui_saves_neural_weight_product():
+    maui_model = Maui(n_hidden=[10], n_latent=2, epochs=1)
+    z = maui_model.fit_transform({"d1": df1, "d2": df2})
+    nwp = maui_model.get_neural_weight_product()
+    assert nwp is not None
+    assert hasattr(maui_model, "nwp_")
+
+    print(maui_model.encoder.summary())
+
+    w1 = maui_model.encoder.layers[2].get_weights()[0]
+    w2 = maui_model.encoder.layers[3].get_weights()[0]
+
+    nwp_11 = np.dot(w1[0, :], w2[:, 0])
+    assert np.allclose(nwp_11, nwp.iloc[0, 0])
+
+
+def test_maui_updates_neural_weight_product_when_training():
+    maui_model = Maui(n_hidden=[10], n_latent=2, epochs=1)
+
+    z_before = maui_model.fit_transform({"d1": df1, "d2": df2})
+    nwp_before_fine_tuning = maui_model.get_neural_weight_product()
+
+    maui_model.fine_tune({"d1": df1, "d2": df2})
+    z_after = maui_model.transform({"d1": df1, "d2": df2})
+    nwp_after_fine_tuning = maui_model.get_neural_weight_product()
+
+    assert not np.allclose(z_before, z_after)
+    assert not np.allclose(nwp_before_fine_tuning, nwp_after_fine_tuning)
+
+
 def test_maui_clusters_with_single_k():
     maui_model = Maui(n_hidden=[10], n_latent=2, epochs=1)
     maui_model.z_ = pd.DataFrame(
@@ -120,7 +153,7 @@ def test_maui_clusters_picks_optimal_k_by_ami():
         maui_model.cluster(
             ami_y=the_y, optimal_k_range=[1, 2, 3]
         )  # the second trial is k=2
-
+        print(maui_model.kmeans_scores)
         assert maui_model.optimal_k_ == 2
 
 
@@ -200,6 +233,7 @@ def test_maui_computes_roc_and_auc():
 
 
 def test_maui_clusters_only_samples_in_y_index_when_optimizing():
+    np.random.seed(0)
     maui_model = Maui(n_hidden=[10], n_latent=2, epochs=1)
     maui_model.z_ = pd.DataFrame(
         np.random.randn(10, 2),
@@ -259,7 +293,7 @@ def test_select_clinical_factors():
         index=[f"sample {i}" for i in range(11)],
     )
 
-    z_clin = maui_model.select_clinical_factors(survival, cox_penalizer=1)
+    z_clin = maui_model.select_clinical_factors(survival, cox_penalizer=1, alpha=0.1)
     assert "LF0" in z_clin.columns
     assert "LF5" not in z_clin.columns
 
@@ -308,12 +342,12 @@ def test_maui_computes_harrells_c():
         duration_column="duration",
         observed_column="observed",
         cox_penalties=[0.1, 1, 10, 100, 1000, 10000],
-        cv_folds=2,
-        sel_clin_alpha=0.05,
+        cv_folds=3,
+        sel_clin_alpha=0.1,
         sel_clin_penalty=1,
     )
     print(cs)
-    assert np.allclose(cs, [0.8, 0.8])
+    assert np.allclose(cs, [0.5, 0.8, 0.5], atol=0.05)
 
 
 def test_maui_produces_same_prediction_when_run_twice():
@@ -494,13 +528,14 @@ def test_maui_merge_latent_factors_complains_if_unknown_merge_by():
             distance_in="xxx", distance_metric="euclidean"
         )
 
+
 def test_maui_can_save_to_folder():
     maui_model = Maui(n_hidden=[10], n_latent=2, epochs=1)
     maui_model = maui_model.fit({"d1": df1, "d2": df2})
     with tempfile.TemporaryDirectory() as tmpdirname:
         maui_model.save(tmpdirname)
-        assert os.path.isfile(os.path.join(tmpdirname, 'maui_weights.h5'))
-        assert os.path.isfile(os.path.join(tmpdirname, 'maui_args.json'))
+        assert os.path.isfile(os.path.join(tmpdirname, "maui_weights.h5"))
+        assert os.path.isfile(os.path.join(tmpdirname, "maui_args.json"))
 
 
 def test_maui_can_load_from_folder():
@@ -511,9 +546,96 @@ def test_maui_can_load_from_folder():
         maui_model_from_disk = Maui.load(tmpdirname)
 
     assert maui_model_from_disk.n_latent == maui_model.n_latent
-    assert np.allclose(maui_model.vae.get_weights()[0],
-        maui_model_from_disk.vae.get_weights()[0])
+    assert np.allclose(
+        maui_model.vae.get_weights()[0], maui_model_from_disk.vae.get_weights()[0]
+    )
     assert np.allclose(
         maui_model.transform({"d1": df1, "d2": df2}),
-        maui_model_from_disk.transform({"d1": df1, "d2": df2})
-        )
+        maui_model_from_disk.transform({"d1": df1, "d2": df2}),
+    )
+
+
+def test_maui_can_print_verbose_training(capsys):
+    maui_model = Maui(n_hidden=[10], n_latent=2, epochs=1)
+    maui_model = maui_model.fit({"d1": df1, "d2": df2})
+
+    stdout, stderr = capsys.readouterr()
+    assert stdout == ""
+
+    maui_model = Maui(n_hidden=[10], n_latent=2, epochs=1, verbose=1)
+    maui_model = maui_model.fit({"d1": df1, "d2": df2})
+
+    stdout, stderr = capsys.readouterr()
+    assert "Epoch" in stdout
+
+
+def test_maui_model_makes_2_layer_vae():
+    maui_model = Maui(n_hidden=[10], n_latent=2, epochs=1, input_dim=10)
+    layers_names = [l.name for l in maui_model.vae.layers]
+
+    assert "hidden_dim_0_mean" in layers_names
+    assert "latent_mean" in layers_names
+    assert "decode_hidden_0" in layers_names
+    assert "reconstruction" in layers_names
+
+    assert "decode_hidden_1" not in layers_names
+
+
+def test_maui_model_makes_one_layer_vae():
+    maui_model = Maui(n_hidden=[], n_latent=2, epochs=1, input_dim=10)
+    layers_names = [l.name for l in maui_model.vae.layers]
+
+    print(layers_names)
+
+    assert layers_names[-1] == "reconstruction"
+
+    assert not any(
+        "decode_hidden" in name for name in layers_names
+    ), "Has a decode hidden..."
+    assert not any("hidden_dim" in name for name in layers_names), "Has a hidden dim..."
+
+
+def test_maui_model_validates_feature_names_on_predict_after_fit():
+    maui_model = Maui(n_hidden=[10], n_latent=2, epochs=1)
+    maui_model.fit({"d1": df1, "d2": df2})
+
+    z = maui_model.transform({"d1": df1, "d2": df2})
+
+    df1_wrong_features = df1.reindex(df1.index[: len(df1.index) - 1])
+    with pytest.raises(ValueError):
+        z = maui_model.transform({"df1": df1_wrong_features, "df2": df2})
+
+
+def test_maui_model_saves_feature_names_to_disk():
+    maui_model = Maui(n_hidden=[10], n_latent=2, epochs=1)
+    maui_model = maui_model.fit({"d1": df1, "d2": df2})
+    with tempfile.TemporaryDirectory() as tmpdirname:
+        maui_model.save(tmpdirname)
+        maui_model_from_disk = Maui.load(tmpdirname)
+    assert maui_model.feature_names == maui_model_from_disk.feature_names
+
+
+def test_maui_model_loads_model_without_feature_names_from_disk_and_warns():
+    maui_model = Maui(n_hidden=[10], n_latent=2, epochs=1)
+    maui_model = maui_model.fit({"d1": df1, "d2": df2})
+    with tempfile.TemporaryDirectory() as tmpdirname:
+        maui_model.save(tmpdirname)
+        os.remove(os.path.join(tmpdirname, "maui_feature_names.txt"))
+        with pytest.warns(MauiWarning):
+            maui_model_from_disk = Maui.load(tmpdirname)
+        assert maui_model_from_disk.feature_names is None
+
+
+def test_maui_can_fine_tune():
+    maui_model = Maui(n_hidden=[], n_latent=2, epochs=1)
+    maui_model = maui_model.fit({"d1": df1, "d2": df2})
+    maui_model.fine_tune({"d1": df1, "d2": df2}, epochs=1)
+
+
+def test_maui_complains_if_fine_tune_with_wrong_features():
+    maui_model = Maui(n_hidden=[], n_latent=2, epochs=1)
+    maui_model.fit({"d1": df1, "d2": df2})
+
+    df1_wrong_features = df1.reindex(df1.index[: len(df1.index) - 1])
+    with pytest.raises(ValueError):
+        z = maui_model.fine_tune({"df1": df1_wrong_features, "df2": df2})
